"""Testing fmu-ensemble."""
# pylint: disable=protected-access

import datetime
import logging
import os
import shutil
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import yaml
from dateutil.relativedelta import relativedelta
from fmu import ensemble
from resdata.summary import Summary

from .test_ensembleset import symlink_iter

try:
    SKIP_FMU_TOOLS = False
    from fmu.tools import volumetrics
except ImportError:
    SKIP_FMU_TOOLS = True

logger = logging.getLogger(__name__)


def test_single_realization(tmpdir):
    """Test internalization of properties pertaining
    to single realizations"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    tmpdir.chdir()

    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    assert os.path.isabs(real.runpath())
    assert os.path.exists(real.runpath())

    assert len(real.files) == 4
    assert "parameters.txt" in real.data
    assert isinstance(real.parameters["RMS_SEED"], int)
    assert real.parameters["RMS_SEED"] == 422851785
    assert isinstance(real.parameters["MULTFLT_F1"], float)
    assert isinstance(
        real.load_txt("parameters.txt", convert_numeric=False, force_reread=True)[
            "RMS_SEED"
        ],
        str,
    )
    # We have rerun load_txt on parameters, but file count
    # should not increase:
    assert len(real.files) == 4

    with pytest.raises(IOError):
        real.load_txt("nonexistingfile.txt")

    # Load more data from text files:
    assert "NPV" in real.load_txt("outputs.txt")
    assert len(real.files) == 5
    assert "outputs.txt" in real.data
    assert "top_structure" in real.data["outputs.txt"]

    # STATUS file
    status = real.get_df("STATUS")
    assert not status.empty
    assert isinstance(status, pd.DataFrame)
    assert "ECLIPSE" in status.loc[49, "FORWARD_MODEL"]
    assert int(status.loc[49, "DURATION"]) == 141

    # CSV file loading
    vol_df = real.load_csv("share/results/volumes/simulator_volume_fipnum.csv")
    assert len(real.files) == 6
    assert isinstance(vol_df, pd.DataFrame)
    assert vol_df["STOIIP_TOTAL"].sum() > 0

    # Test later retrieval of cached data:
    vol_df2 = real.get_df("share/results/volumes/simulator_volume_fipnum.csv")
    assert vol_df2["STOIIP_TOTAL"].sum() > 0

    # Test scalar import
    assert "OK" in real.keys()  # Imported in __init__
    assert real["OK"] == "All jobs complete 22:47:54"
    # NB: Trailing whitespace from the OK-file is removed.
    assert isinstance(real["OK"], str)

    # Check that we can "reimport" the OK file
    real.load_scalar("OK", force_reread=True)
    assert "OK" in real.keys()  # Imported in __init__
    assert real["OK"] == "All jobs complete 22:47:54"
    assert isinstance(real["OK"], str)
    assert len(real.files[real.files.LOCALPATH == "OK"]) == 1

    real.load_scalar("npv.txt")
    assert real.get_df("npv.txt") == 3444
    assert real["npv.txt"] == 3444
    assert isinstance(real.data["npv.txt"], (int, np.integer))
    assert "npv.txt" in real.files.LOCALPATH.values
    assert real.files[real.files.LOCALPATH == "npv.txt"]["FILETYPE"].values[0] == "txt"

    real.load_scalar("emptyscalarfile")
    # Activate this test when filter() is merged:
    # assert real.contains('emptyfile')
    assert "emptyscalarfile" in real.data
    assert isinstance(real["emptyscalarfile"], str)
    assert "emptyscalarfile" in real.files["LOCALPATH"].values

    # Check that FULLPATH always has absolute paths
    assert all([os.path.isabs(x) for x in real.files["FULLPATH"]])

    with pytest.raises(IOError):
        real.load_scalar("notexisting.txt")

    # Test internal storage:
    localpath = "share/results/volumes/simulator_volume_fipnum.csv"
    assert localpath in real.data
    assert isinstance(real.get_df(localpath), pd.DataFrame)
    assert isinstance(real.get_df("parameters.txt"), dict)
    assert isinstance(real.get_df("outputs.txt"), dict)

    # Test shortcuts to the internal datastore
    assert isinstance(real.get_df("simulator_volume_fipnum.csv"), pd.DataFrame)
    # test without extension:
    assert isinstance(
        real.get_df("share/results/volumes/" + "simulator_volume_fipnum"), pd.DataFrame
    )
    assert isinstance(real.get_df("parameters"), dict)
    # test basename and no extension:
    assert isinstance(real.get_df("simulator_volume_fipnum"), pd.DataFrame)

    # Some CSV files might already contain the REAL column, this is not allowed
    foo_df = pd.DataFrame(columns=["REAL", "FOOBAR"], data=[[0, 1], [2, 3]])
    foo_df.to_csv(os.path.join(realdir, "foo-real.csv"), index=False)
    real.load_csv("foo-real.csv")  # A warning will be issued.
    assert "REAL" not in real.get_df("foo-real")
    assert "FOOBAR" in real.get_df("foo-real")

    with pytest.raises((ValueError, KeyError)):
        real.get_df("notexisting.csv")

    # Test __delitem__()
    keycount = len(real.keys())
    del real["parameters.txt"]
    assert len(real.keys()) == keycount - 1

    # At realization level, wrong filenames should throw exceptions,
    # at ensemble level it is fine.
    with pytest.raises(IOError):
        real.load_csv("bogus.csv")


def test_status_load(tmpdir):
    """Test loading of STATUS file with different errors in them

    These files are custom text files, and can have stray error
    messages in them. Robustness (i.e. no crash) is more
    important than best-effort parsing. Better parsing
    is left for issue #12
    """

    # Mock a realization:
    tmpdir.join("realization-0").mkdir()

    # Test with some selected STATUS files:
    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    assert real.get_df("STATUS").empty

    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("\n")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    assert real.get_df("STATUS").empty

    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("foo bar bogus com\n")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    assert real.get_df("STATUS").empty

    # Two sucessfull jobs
    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("first line always ignored\n")
        status_fh.write("INCLUDE_PC                      : 12:40:55 .... 12:40:55  \n")
        status_fh.write("ECLIPSE100_2014.2               : 12:40:55 .... 12:43:16\n")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    status = real.get_df("STATUS")
    assert len(status) == 2
    assert "FORWARD_MODEL" in status
    assert "STARTTIME" in status
    assert "ENDTIME" in status
    assert "DURATION" in status
    assert (status["DURATION"].values == [0, 141]).all()  # in seconds

    # Two sucessfull jobs, but with error string
    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("first line always ignored\n")
        status_fh.write("INCLUDE_PC                      : 12:40:55 .... 12:40:55  \n")
        status_fh.write(
            "ECLIPSE100_2014.2               : 12:40:55 "
            ".... 12:43:16 SOMEERRORSTRING_FOO\n"
        )
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    status = real.get_df("STATUS")
    assert len(status) == 2
    assert "FORWARD_MODEL" in status
    assert (status["DURATION"].values == [0, 141]).all()  # in seconds
    assert status["errorstring"].values[1] == "SOMEERRORSTRING_FOO"

    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("first line always ignored\n")
        status_fh.write("INCLUDE_PC                      : 12:40:55 .... 12:40:55  \n")
        status_fh.write("ECLIPSE100_2014.2               : 12:40:55 ....")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    status = real.get_df("STATUS")
    assert len(status) == 2
    assert "FORWARD_MODEL" in status
    assert status["DURATION"].values[0] == 0  # in seconds
    assert np.isnan(status["DURATION"].values[1])

    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("first line always ignored\n")
        status_fh.write("INCLUDE_PC                      : 12:40:55 .... 12:40:55  \n")
        status_fh.write("error message: something went really wrong")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    status = real.get_df("STATUS")
    assert len(status) == 2
    assert "FORWARD_MODEL" in status
    assert status["DURATION"].values[0] == 0  # in seconds
    # NB: The current code is not able to pick this error string

    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("first line always ignored\n")
        status_fh.write("INCLUDE_PC                      : 12:XX:55 .... 12:40:55  \n")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    status = real.get_df("STATUS")
    assert len(status) == 1
    assert "FORWARD_MODEL" in status
    assert np.isnan(status["DURATION"].values[0])

    with open(str(tmpdir.join("realization-0/STATUS")), "w") as status_fh:
        status_fh.write("first line always ignored\n")
        status_fh.write("INCLUDE_PC                      : 12:40.... 12:40:55  \n")
    real = ensemble.ScratchRealization(str(tmpdir.join("realization-0")))
    status = real.get_df("STATUS")
    assert len(status) == 1
    assert "FORWARD_MODEL" in status
    assert np.isnan(status["DURATION"].values[0])  # Unsupported time syntax


def test_batch():
    """Test batch processing at time of object initialization"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(
        realdir,
        batch=[
            {"load_scalar": {"localpath": "npv.txt"}},
            {"load_smry": {"column_keys": "FOPT", "time_index": "yearly"}},
            {"load_smry": {"column_keys": "*", "time_index": "daily"}},
            {"load_smry": {"column_keys": "*", "time_index": "weekly"}},
            {"illegal-ignoreme": {}},
        ],
    )
    assert real.get_df("npv.txt") == 3444
    assert len(real.get_df("unsmry--daily")["FOPR"]) > 2
    assert len(real.get_df("unsmry--yearly")["FOPT"]) > 2
    assert len(real.get_df("unsmry--weekly")["FOPT"]) > 2


def test_volumetric_rates():
    """Test computation of volumetric rates from cumulative vectors"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    # Should work without prior internalization:
    cum_df = real.get_smry(column_keys=["F*T", "W*T*"], time_index="yearly")
    vol_rate_df = real.get_volumetric_rates(
        column_keys=["F*T", "W*T*"], time_index="yearly"
    )
    assert vol_rate_df.index.name == "DATE"
    assert "FWCR" not in vol_rate_df  # We should not compute FWCT..
    assert "FOPR" in vol_rate_df
    assert "FWPR" in vol_rate_df

    # Test that computed rates can be summed up to cumulative at end:
    assert vol_rate_df["FOPR"].sum() == cum_df["FOPT"].iloc[-1]
    assert vol_rate_df["FGPR"].sum() == cum_df["FGPT"].iloc[-1]
    assert vol_rate_df["FWPR"].sum() == cum_df["FWPT"].iloc[-1]

    # Since rates are valid forwards in time, the last
    # row should have a zero, since that is the final simulated
    # date for the cumulative vector
    assert vol_rate_df["FOPR"].iloc[-1] == 0

    # Check that we allow cumulative allocated vectors:
    cumvecs = real.get_volumetric_rates(column_keys=["F*TH", "W*TH*"])
    assert not cumvecs.empty
    assert "FOPRH" in cumvecs
    assert "WOPRH:OP_1" in cumvecs

    assert real.get_volumetric_rates(column_keys="FOOBAR").empty
    assert real.get_volumetric_rates(column_keys=["FOOBAR"]).empty
    assert real.get_volumetric_rates(column_keys={}).empty

    with pytest.raises(ValueError):
        real.get_volumetric_rates(column_keys="FOPT", time_index="bogus")

    mcum = real.get_smry(column_keys="FOPT", time_index="monthly")
    dmcum = real.get_volumetric_rates(column_keys="FOPT", time_index="monthly")
    assert dmcum["FOPR"].sum() == mcum["FOPT"].iloc[-1]

    # Pick 10 **random** dates to get the volumetric rates between:
    daily_dates = real.get_smry_dates(freq="daily", normalize=False)
    subset_dates = np.random.choice(daily_dates, size=10, replace=False)
    subset_dates.sort()
    dcum = real.get_smry(column_keys="FOPT", time_index=subset_dates)
    ddcum = real.get_volumetric_rates(column_keys="FOPT", time_index=subset_dates)
    assert ddcum["FOPR"].iloc[-1] == 0

    # We are probably neither at the start or at the end of the production
    # interval.
    cumulative_error = ddcum["FOPR"].sum() - (
        dcum["FOPT"].loc[subset_dates[-1]] - dcum["FOPT"].loc[subset_dates[0]]
    )

    # Give some slack, we might have done a lot of interpolation
    # here.
    assert cumulative_error / ddcum["FOPR"].sum() < 0.000001

    # Test the time_unit feature.
    vol_rate_days = real.get_volumetric_rates(
        column_keys=["F*T"], time_index="yearly", time_unit="days"
    )
    vol_rate_months = real.get_volumetric_rates(
        column_keys=["F*T"], time_index="yearly", time_unit="months"
    )
    vol_rate_years = real.get_volumetric_rates(
        column_keys=["F*T"], time_index="yearly", time_unit="years"
    )

    # Sample test on correctness.
    # Fine-accuracy (wrt leap days) is not tested here:
    assert vol_rate_days["FWIR"].iloc[0] * 27.9 < vol_rate_months["FWIR"].iloc[0]
    assert vol_rate_days["FWIR"].iloc[0] * 31.1 > vol_rate_months["FWIR"].iloc[0]
    assert vol_rate_months["FWIR"].iloc[0] * 12 == pytest.approx(
        vol_rate_years["FWIR"].iloc[0]
    )

    assert vol_rate_days["FWIR"].iloc[0] * 364.9 < vol_rate_years["FWIR"].iloc[0]
    assert vol_rate_days["FWIR"].iloc[0] * 366.1 > vol_rate_years["FWIR"].iloc[0]

    with pytest.raises(ValueError):
        real.get_volumetric_rates(
            column_keys=["F*T"], time_index="yearly", time_unit="bogus"
        )

    # Try with the random dates
    dayscum = real.get_volumetric_rates(
        column_keys="FOPT", time_index=subset_dates, time_unit="days"
    )
    assert all(np.isfinite(dayscum["FOPR"]))
    diffdays = pd.DataFrame(pd.to_datetime(dayscum.index)).diff().shift(-1)
    dayscum["DIFFDAYS"] = [x.days for x in diffdays["DATE"]]
    # Calculate cumulative production from the computed volumetric daily rates:
    dayscum["FOPRcum"] = dayscum["FOPR"] * dayscum["DIFFDAYS"]
    # Check that this sum is equal to FOPT between first and last date:
    assert dayscum["FOPRcum"].sum() == pytest.approx(dcum["FOPT"][-1] - dcum["FOPT"][0])
    # (here we could catch an error in case we don't support leap days)

    # Monthly rates between the random dates:
    monthlyrates = real.get_volumetric_rates(
        column_keys="FOPT", time_index=subset_dates, time_unit="months"
    )
    assert all(np.isfinite(monthlyrates["FOPR"]))

    # Total number of months in production period
    delta = relativedelta(vol_rate_days.index[-1], vol_rate_days.index[0])
    months = delta.years * 12 + delta.months
    tworows = real.get_volumetric_rates(
        column_keys="FOPT",
        time_index=[vol_rate_days.index[0], vol_rate_days.index[-1]],
        time_unit="months",
    )
    assert tworows["FOPR"].iloc[0] * months == pytest.approx(cum_df["FOPT"].iloc[-1])

    # Check for defaults and error handling:
    assert not real.get_smry(column_keys=None).empty
    assert not real.get_smry(column_keys=[None]).empty
    assert not real.get_smry(column_keys=[None, "WOPT:BOGUS"]).empty
    assert not real.get_smry(column_keys=["WOPT:BOGUS", None]).empty
    column_count = len(real.get_smry())
    # Columns repeatedly asked for should not be added:
    assert len(real.get_smry(column_keys=[None, "FOPT"])) == column_count
    assert len(real.get_smry(column_keys=[None, "FOPT", "FOPT"])) == column_count
    assert real.get_smry(column_keys=["WOPT:BOGUS"]).empty

    assert "FOPT" in real.get_smry(column_keys=["WOPT:BOGUS", "FOPT"])


def test_datenormalization():
    """Test normalization of dates, where
    dates can be ensured to be on dategrid boundaries"""

    # fmu.ensemble.util.normalized_dates is also tested in test_util.py

    # Check that we normalize correctly with get_smry():
    # realization-0 here has its last summary date at 2003-01-02
    testdir = os.path.dirname(os.path.abspath(__file__))
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)
    raw = real.get_smry(column_keys="FOPT", time_index="raw")
    assert str(raw.index[-1]) == "2003-01-02 00:00:00"
    daily = real.get_smry(column_keys="FOPT", time_index="daily")
    assert str(daily.index[-1]) == "2003-01-02"
    monthly = real.get_smry(column_keys="FOPT", time_index="monthly")
    assert str(monthly.index[-1]) == "2003-02-01"
    yearly = real.get_smry(column_keys="FOPT", time_index="yearly")
    assert str(yearly.index[-1]) == "2004-01-01"
    weekly = real.get_smry(column_keys="FOPT", time_index="weekly")
    assert str(weekly.index[-1]) == "2003-01-06"  # First Monday after 2003-01-02
    weekly = real.get_smry(column_keys="FOPT", time_index="W-MON")
    assert str(weekly.index[-1]) == "2003-01-06"  # First Monday after 2003-01-02
    weekly = real.get_smry(column_keys="FOPT", time_index="W-TUE")
    assert str(weekly.index[-1]) == "2003-01-07"  # First Tuesday after 2003-01-02
    weekly = real.get_smry(column_keys="FOPT", time_index="W-THU")
    assert str(weekly.index[-1]) == "2003-01-02"  # First Thursday after 2003-01-02

    # Check that time_index=None and time_index="raw" behaves like default
    raw = real.load_smry(column_keys="FOPT", time_index="raw")
    assert list(real.load_smry(column_keys="FOPT")["FOPT"].values) == list(
        raw["FOPT"].values
    )
    assert list(
        real.load_smry(column_keys="FOPT", time_index=None)["FOPT"].values
    ) == list(raw["FOPT"].values)

    # Check that we get the same correct normalization
    # with load_smry()
    real.load_smry(column_keys="FOPT", time_index="raw")
    assert str(real.get_df("unsmry--raw")["DATE"].iloc[-1]) == "2003-01-02 00:00:00"
    real.load_smry(column_keys="FOPT", time_index="daily")
    assert str(real.get_df("unsmry--daily")["DATE"].iloc[-1]) == "2003-01-02"
    real.load_smry(column_keys="FOPT", time_index="monthly")
    assert str(real.get_df("unsmry--monthly")["DATE"].iloc[-1]) == "2003-02-01"
    real.load_smry(column_keys="FOPT", time_index="yearly")
    assert str(real.get_df("unsmry--yearly")["DATE"].iloc[-1]) == "2004-01-01"
    real.load_smry(column_keys="FOPT", time_index="weekly")
    assert str(real.get_df("unsmry--weekly")["DATE"].iloc[-1]) == "2003-01-06"


def test_singlereal_ecl(tmp="TMP"):
    """Test Eclipse specific functionality for realizations"""

    testdir = os.path.dirname(os.path.abspath(__file__))
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    # Eclipse summary files:
    assert isinstance(real.get_eclsum(), Summary)
    if not os.path.exists(tmp):
        os.mkdir(tmp)
    real.load_smry().to_csv(os.path.join(tmp, "real0smry.csv"), index=False)
    assert real.load_smry().shape == (378, 474)
    # 378 dates, 470 columns + DATE column

    assert real.load_smry(column_keys=["FOP*"])["FOPT"].max() > 6000000
    assert real.get_smryvalues("FOPT")["FOPT"].max() > 6000000

    # get_smry() should be analogue to load_smry(), but it should
    # not modify the internalized dataframes!
    internalized_df = real["unsmry--raw"]
    fresh_df = real.get_smry(column_keys=["G*"])
    assert "GGIR:OP" in fresh_df.columns
    assert "GGIR:OP" not in internalized_df.columns
    # Test that the internalized was not touched:
    assert "GGIR:OP" not in real["unsmry--raw"].columns

    assert "FOPT" in real.get_smry(column_keys=["F*"], time_index="monthly")
    assert "FOPT" in real.get_smry(column_keys="F*", time_index="yearly")
    assert "FOPT" in real.get_smry(column_keys="FOPT", time_index="daily")
    assert "FOPT" in real.get_smry(column_keys="FOPT", time_index="weekly")
    assert "FOPT" in real.get_smry(column_keys="FOPT", time_index="raw")

    # Test date functionality
    assert isinstance(real.get_smry_dates(), list)
    assert isinstance(real.get_smry_dates(freq="last"), list)
    assert isinstance(real.get_smry_dates(freq="last")[0], datetime.date)
    assert len(real.get_smry_dates()) == len(real.get_smry_dates(freq="monthly"))
    monthly = real.get_smry_dates(freq="monthly")
    assert monthly[-1] > monthly[0]  # end date is later than start
    assert len(real.get_smry_dates(freq="yearly")) == 5
    assert len(monthly) == 38
    assert len(real.get_smry_dates(freq="daily")) == 1098

    # Try ISO-date for time_index:
    singledata = real.get_smry(time_index="2000-05-05", column_keys="FOPT")
    assert "FOPT" in singledata
    assert "2000-05-05" in singledata.index

    # start and end should be included:
    assert (
        len(
            real.get_smry_dates(
                start_date="2000-06-05", end_date="2000-06-07", freq="daily"
            )
        )
        == 3
    )
    # No month boundary between start and end, but we
    # should have the starts and ends included
    assert (
        len(
            real.get_smry_dates(
                start_date="2000-06-05", end_date="2000-06-07", freq="monthly"
            )
        )
        == 2
    )
    # Date normalization should be overridden here:
    assert (
        len(
            real.get_smry_dates(
                start_date="2000-06-05",
                end_date="2000-06-07",
                freq="monthly",
                normalize=True,
            )
        )
        == 2
    )

    # Start_date and end_date at the same date should work
    assert len(real.get_smry_dates(start_date="2000-01-01", end_date="2000-01-01")) == 1
    assert (
        len(
            real.get_smry_dates(
                start_date="2000-01-01", end_date="2000-01-01", normalize=True
            )
        )
        == 1
    )

    # Check that we can go way outside the smry daterange:
    assert (
        len(
            real.get_smry_dates(
                start_date="1978-01-01", end_date="2030-01-01", freq="yearly"
            )
        )
        == 53
    )
    assert (
        len(
            real.get_smry_dates(
                start_date="1978-01-01",
                end_date="2030-01-01",
                freq="yearly",
                normalize=True,
            )
        )
        == 53
    )

    assert (
        len(
            real.get_smry_dates(
                start_date="2000-06-05",
                end_date="2000-06-07",
                freq="raw",
                normalize=True,
            )
        )
        == 2
    )
    assert (
        len(
            real.get_smry_dates(
                start_date="2000-06-05",
                end_date="2000-06-07",
                freq="raw",
                normalize=False,
            )
        )
        == 2
    )

    # Test caching/internalization of summary files

    # This should be false, since only the full localpath is in keys():
    assert "unsmry--raw.csv" not in real.keys()
    assert "share/results/tables/unsmry--raw.csv" in real.keys()
    assert "FOPT" in real["unsmry--raw"]
    with pytest.raises((ValueError, KeyError)):
        # This does not exist before we have asked for it
        # pylint: disable=pointless-statement
        "FOPT" in real["unsmry--yearly"]


def test_can_import_summary_files_beyond_2262(tmpdir, monkeypatch):
    """Pandas is/has been eager to use datetime64[ns] which overflows in year 2262,
    ensure this limitation is sufficiently worked around."""
    monkeypatch.chdir(tmpdir)
    res_sum = Summary.from_pandas(
        "TESTCASE",
        pd.DataFrame(
            [
                {"DATE": datetime.date(2000, 1, 1), "FPR": 200},
                {"DATE": datetime.date(2263, 1, 1), "FPR": 1},
            ]
        ).set_index("DATE"),
    )
    runpath = "realization-0/iter-0"
    Path(runpath).mkdir(parents=True)
    os.chdir(runpath)
    # fwrite() can only write to cwd
    Summary.fwrite(res_sum)
    os.chdir(tmpdir)

    real = ensemble.ScratchRealization(runpath)
    real.find_files("TESTCASE.UNSMRY")
    for time_index in ["raw", "monthly", "yearly"]:
        assert "2263-01-01" in str(
            real.get_smry(column_keys="*", time_index=time_index)
        )
    with pytest.raises(ValueError):
        real.get_smry(column_keys="*", time_index="weekly")


def test_independent_realization(tmp="TMP"):
    """Test what we are able to load a single Eclipse run
    that might have nothing to do with FMU"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    datadir = os.path.join(testdir, "data")
    tmpdir = os.path.join(datadir, tmp)
    if os.path.exists(tmpdir):
        shutil.rmtree(tmpdir)
    os.mkdir(tmpdir)
    # Let the directory contain only the UNSMRY and SMSPEC file
    shutil.copyfile(
        os.path.join(
            datadir,
            "testensemble-reek001/realization-2/iter-0/eclipse/"
            + "model/2_R001_REEK-2.UNSMRY",
        ),
        os.path.join(tmpdir, "2_R001_REEK-2.UNSMRY"),
    )
    shutil.copyfile(
        os.path.join(
            datadir,
            "testensemble-reek001/realization-2/iter-0/eclipse/"
            + "model/2_R001_REEK-2.SMSPEC",
        ),
        os.path.join(tmpdir, "2_R001_REEK-2.SMSPEC"),
    )

    # This should not fail, but with a nice constructive warning to the user hinting
    # about the solution
    empty = ensemble.ScratchRealization(tmpdir)
    assert not empty.index  # The index is None in such realizations.

    # This is how it must be done:
    real = ensemble.ScratchRealization(tmpdir, index="999")
    assert real.index == 999

    # No auto-discovery here:
    assert real.get_smry().empty

    # Explicit discovery:
    real.find_files("*UNSMRY")
    assert not real.get_smry().empty

    # However, we can do something with an undefined index:
    noindex = ensemble.ScratchRealization(tmpdir)
    noindex.find_files("*UNSMRY")
    assert not real.get_smry().empty

    shutil.rmtree(tmpdir)


def test_filesystem_changes():
    """Test loading of sparse realization (random data missing)

    Performed by filesystem manipulations from the original realizations.
    Clean up from previous runs are attempted, and also done when it finishes.
    (after a failed test run, filesystem is tainted)
    """

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    datadir = testdir + "/data"
    tmpensname = ".deleteme_ens"
    # Clean up earlier failed runs:
    if os.path.exists(datadir + "/" + tmpensname):
        shutil.rmtree(datadir + "/" + tmpensname, ignore_errors=True)
    os.mkdir(datadir + "/" + tmpensname)
    shutil.copytree(
        datadir + "/testensemble-reek001/realization-0",
        datadir + "/" + tmpensname + "/realization-0",
    )

    realdir = datadir + "/" + tmpensname + "/realization-0/iter-0"

    # Load untainted realization, nothing bad should happen
    real = ensemble.ScratchRealization(realdir)

    # Remove SMSPEC file and reload:
    shutil.move(
        realdir + "/eclipse/model/2_R001_REEK-0.SMSPEC",
        realdir + "/eclipse/model/2_R001_REEK-0.SMSPEC-FOOO",
    )
    real = ensemble.ScratchRealization(realdir)  # this should go fine
    # This should just return None. Logging info is okay.
    assert real.get_eclsum() is None
    # This should return None
    assert real.get_smry_dates() is None
    # This should return empty dataframe:
    assert isinstance(real.load_smry(), pd.DataFrame)
    assert real.load_smry().empty

    assert isinstance(real.get_smry(), pd.DataFrame)
    assert real.get_smry().empty

    # Also move away UNSMRY and redo:
    shutil.move(
        realdir + "/eclipse/model/2_R001_REEK-0.UNSMRY",
        realdir + "/eclipse/model/2_R001_REEK-0.UNSMRY-FOOO",
    )
    real = ensemble.ScratchRealization(realdir)  # this should go fine
    # This should just return None
    assert real.get_eclsum() is None
    # This should return None
    assert real.get_smry_dates() is None
    # This should return empty dataframe:
    assert isinstance(real.load_smry(), pd.DataFrame)
    assert real.load_smry().empty

    # Reinstate summary data:
    shutil.move(
        realdir + "/eclipse/model/2_R001_REEK-0.UNSMRY-FOOO",
        realdir + "/eclipse/model/2_R001_REEK-0.UNSMRY",
    )
    shutil.move(
        realdir + "/eclipse/model/2_R001_REEK-0.SMSPEC-FOOO",
        realdir + "/eclipse/model/2_R001_REEK-0.SMSPEC",
    )

    # Remove jobs.json, this file should not be critical
    # but the status dataframe should have less information
    statuscolumnswithjson = len(real.get_df("STATUS").columns)
    os.remove(realdir + "/jobs.json")
    real = ensemble.ScratchRealization(realdir)  # this should go fine

    statuscolumnswithoutjson = len(real.get_df("STATUS").columns)
    assert statuscolumnswithoutjson > 0
    # Check that some STATUS info is missing.
    assert statuscolumnswithoutjson < statuscolumnswithjson

    # Remove parameters.txt
    shutil.move(realdir + "/parameters.txt", realdir + "/parameters.text")
    real = ensemble.ScratchRealization(realdir)
    # Should not fail

    # Move it back so the realization is valid again
    shutil.move(realdir + "/parameters.text", realdir + "/parameters.txt")

    # Remove STATUS altogether:
    shutil.move(realdir + "/STATUS", realdir + "/MOVEDSTATUS")
    real = ensemble.ScratchRealization(realdir)
    # Should not fail

    # Try with an empty STATUS file:
    fhandle = open(realdir + "/STATUS", "w")
    fhandle.close()
    real = ensemble.ScratchRealization(realdir)
    assert real.get_df("STATUS").empty
    # This demonstrates we can fool the Realization object, and
    # should perhaps leads to relaxation of the requirement..

    # Try with a STATUS file with error message on first job
    # the situation where there is one successful job.
    fhandle = open(realdir + "/STATUS", "w")
    fhandle.write(
        (
            "Current host                    : st-rst16-02-03/x86_64  "
            "file-server:10.14.10.238\n"
            "LSF JOBID: not running LSF\n"
            "COPY_FILE                       : 20:58:57 .... 20:59:00   "
            "EXIT: 1/Executable: /project/res/komodo/2018.02/root/etc/ERT/"
            "Config/jobs/util/script/copy_file.py failed with exit code: 1\n"
        )
    )
    fhandle.close()
    real = ensemble.ScratchRealization(realdir)
    # When issue 37 is resolved, update this to 1 and check the
    # error message is picked up.
    assert len(real.get_df("STATUS")) == 1
    fhandle = open(realdir + "/STATUS", "w")
    fhandle.write(
        (
            "Current host                    : st-rst16-02-03/x86_64  "
            "file-server:10.14.10.238\n"
            "LSF JOBID: not running LSF\n"
            "COPY_FILE                       : 20:58:55 .... 20:58:57\n"
            "COPY_FILE                       : 20:58:57 .... 20:59:00  "
            " EXIT: 1/Executable: /project/res/komodo/2018.02/root/etc/ERT/"
            "Config/jobs/util/script/copy_file.py failed with exit code: 1 "
        )
    )
    fhandle.close()
    real = ensemble.ScratchRealization(realdir)
    assert len(real.get_df("STATUS")) == 2
    # Check that we have the error string picked up:
    assert real.get_df("STATUS")["errorstring"].dropna().values[0] == (
        "EXIT: 1/Executable: /project/res/komodo/2018.02/root/"
        "etc/ERT/Config/jobs/util/script/copy_file.py failed with exit code: 1"
    )  # noqa

    # Check that we can move the Eclipse files to another place
    # in the realization dir and still load summary data:
    shutil.move(realdir + "/eclipse", realdir + "/eclipsedir")
    real = ensemble.ScratchRealization(realdir)

    # load_smry() is now the same as no UNSMRY file found,
    # an empty dataframe (and there would be some logging)
    assert real.load_smry().empty

    # Now discover the UNSMRY file explicitly, then load_smry()
    # should work.
    unsmry_file = real.find_files("eclipsedir/model/*.UNSMRY")
    # Non-empty dataframe:
    assert not real.load_smry().empty
    assert len(unsmry_file) == 1
    assert isinstance(unsmry_file, pd.DataFrame)

    # Test having values with spaces in parameters.txt
    # Unquoted valued with spaces will be truncated,
    # quoted valued will be correctly parsed
    # (read_csv(sep='\s+') is the parser)
    param_file = open(realdir + "/parameters.txt", "a")
    param_file.write("FOOBAR 1 2 3 4 5 6\n")
    param_file.write('FOOSPACES "1 2 3 4 5 6"\n')
    param_file.close()

    real = ensemble.ScratchRealization(realdir)
    assert real.parameters["FOOBAR"] == 1
    assert real.parameters["FOOSPACES"] == "1 2 3 4 5 6"

    # Clean up when finished. This often fails on network drives..
    shutil.rmtree(datadir + "/" + tmpensname, ignore_errors=True)


def test_apply(tmpdir):
    """
    Test the callback functionality
    """
    testdir = os.path.dirname(os.path.abspath(__file__))
    tmpdir.chdir()

    symlink_iter(os.path.join(testdir, "data/testensemble-reek001"), "iter-0")
    realdir = "realization-0/iter-0"
    real = ensemble.ScratchRealization(realdir)

    def ex_func1():
        """Example constant function"""
        return pd.DataFrame(
            index=["1", "2"], columns=["foo", "bar"], data=[[1, 2], [3, 4]]
        )

    result = real.apply(ex_func1)
    assert isinstance(result, pd.DataFrame)

    # Apply and store the result:
    real.apply(ex_func1, localpath="df-1234")
    internalized_result = real.get_df("df-1234")
    assert isinstance(internalized_result, pd.DataFrame)
    assert (result == internalized_result).all().all()

    # Check that the submitted function can utilize data from **kwargs
    def ex_func2(kwargs):
        """Example function using kwargs"""
        arg = kwargs["foo"]
        return pd.DataFrame(
            index=["1", "2"], columns=["foo", "bar"], data=[[arg, arg], [arg, arg]]
        )

    result2 = real.apply(ex_func2, foo="bar")
    assert result2.iloc[0, 0] == "bar"

    # We require applied function to return only DataFrames.
    def scalar_func():
        """Dummy scalar function"""
        return 1

    with pytest.raises(ValueError):
        real.apply(scalar_func)

    # The applied function should have access to the realization object:
    def real_func(kwargs):
        """Example function that accesses the realization object"""
        return pd.DataFrame(
            index=[0], columns=["path"], data=kwargs["realization"].runpath()
        )

    origpath = real.apply(real_func)
    assert os.path.exists(origpath.iloc[0, 0])

    # Do not allow supplying the realization object to apply:
    with pytest.raises(ValueError):
        real.apply(real_func, realization="foo")

    if SKIP_FMU_TOOLS:
        return
    # Test if we can wrap the volumetrics-parser in fmu.tools:
    # It cannot be applied directly, as we need to combine the
    # realization's root directory with the relative path coming in:

    def rms_vol2df(kwargs):
        """Small function that is to be supplied to .apply()"""
        return volumetrics.rmsvolumetrics_txt2df(
            os.path.join(kwargs["realization"].runpath(), kwargs["filename"])
        )

    rmsvol_df = real.apply(
        rms_vol2df, filename="share/results/volumes/" + "geogrid_vol_oil_1.txt"
    )
    assert rmsvol_df["STOIIP_OIL"].sum() > 0

    # Also try internalization:
    real.apply(
        rms_vol2df,
        filename="share/results/volumes/" + "geogrid_vol_oil_1.txt",
        localpath="share/results/volumes/geogrid--oil.csv",
    )
    assert real.get_df("geogrid--oil")["STOIIP_OIL"].sum() > 0

    # Run rms_vol2df in batch when initializing:
    real = ensemble.ScratchRealization(
        realdir,
        batch=[
            {
                "apply": {
                    "callback": rms_vol2df,
                    "filename": "share/results/volumes/" + "geogrid_vol_oil_1.txt",
                    "localpath": "share/results/volumes/geogrid--oil.csv",
                }
            }
        ],
    )
    assert real.get_df("geogrid--oil")["STOIIP_OIL"].sum() > 0


def test_drop():
    """Test the drop functionality, where can delete
    parts of internalized data"""
    testdir = os.path.dirname(os.path.abspath(__file__))
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    parametercount = len(real.parameters)
    real.drop("parameters", key="RMS_SEED")
    assert len(real.parameters) == parametercount - 1

    real.drop("parameters", keys=["L_1GO", "E_1GO"])
    assert len(real.parameters) == parametercount - 3

    real.drop("parameters", key="notexistingkey")
    # This will go unnoticed
    assert len(real.parameters) == parametercount - 3

    real.load_smry(column_keys="FOPT", time_index="monthly")
    datecount = len(real.get_df("unsmry--monthly"))
    real.drop("unsmry--monthly", rowcontains="2000-01-01")
    assert len(real.get_df("unsmry--monthly")) == datecount - 1

    real.drop("parameters")
    assert "parameters.txt" not in real.keys()


def test_find_files_comps():
    """Test the more exotic features of find_files

    Components extracted from filenames.
    """

    testdir = os.path.dirname(os.path.abspath(__file__))
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    # Make some filenames we can later "find", including some problematic ones.
    findable_files = [
        "foo--bar--com.gri",
        "foo-bar--com.gri",
        "foo---bar--com.gri",
        "--bar--.gri",
    ]
    for filename in findable_files:
        with open(os.path.join(realdir, filename), "w") as fileh:
            fileh.write("baah")

    real.find_files("*.gri")

    files_df = real.files.set_index("BASENAME")
    assert "COMP0" not in real.files  # We are 1-based, not zero.
    assert "COMP1" in real.files
    assert "COMP2" in real.files
    assert "COMP3" in real.files
    assert "COMP4" not in real.files

    assert files_df.loc["foo--bar--com.gri"]["COMP1"] == "foo"
    assert files_df.loc["foo--bar--com.gri"]["COMP2"] == "bar"
    assert files_df.loc["foo--bar--com.gri"]["COMP3"] == "com"
    assert files_df.loc["foo-bar--com.gri"]["COMP1"] == "foo-bar"
    assert files_df.loc["foo-bar--com.gri"]["COMP2"] == "com"
    assert files_df.loc["foo---bar--com.gri"]["COMP1"] == "foo"
    assert files_df.loc["foo---bar--com.gri"]["COMP2"] == "-bar"
    assert files_df.loc["foo---bar--com.gri"]["COMP3"] == "com"
    assert files_df.loc["--bar--.gri"]["COMP1"] == ""
    assert files_df.loc["--bar--.gri"]["COMP2"] == "bar"
    assert files_df.loc["--bar--.gri"]["COMP3"] == ""

    # Cleanup
    for filename in findable_files:
        if os.path.exists(os.path.join(realdir, filename)):
            os.unlink(os.path.join(realdir, filename))


def test_find_files_yml():
    """Test the more exotic features of find_files

    Meta-data in yaml files.
    """
    testdir = os.path.dirname(os.path.abspath(__file__))
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    # Setup example files with some yaml data:
    findable_files = ["grid1.gri", "grid2.gri"]
    for filename in findable_files:
        with open(os.path.join(realdir, filename), "w") as fileh:
            fileh.write("baah")
        yamlfile = "." + filename + ".yml"
        with open(os.path.join(realdir, yamlfile), "w") as fileh:
            fileh.write(yaml.dump(dict(a=dict(x=1, y=2), b="bar")))

    # Now find the gri files, and add metadata:
    files_df = real.find_files("*.gri", metayaml=True)

    assert "a--x" in files_df
    assert "a--y" in files_df
    assert "b" in files_df
    assert files_df["b"].unique()[0] == "bar"
    assert files_df["a--x"].astype(int).unique()[0] == 1
    assert files_df["a--y"].astype(int).unique()[0] == 2

    # Cleanup
    for filename in findable_files:
        if os.path.exists(os.path.join(realdir, filename)):
            os.unlink(os.path.join(realdir, filename))
        yamlfile = "." + filename + ".yml"
        if os.path.exists(os.path.join(realdir, yamlfile)):
            os.unlink(os.path.join(realdir, yamlfile))


def test_get_smry_meta():
    """
    Test getting eclsum metadata for single realization.
    """
    testdir = os.path.dirname(os.path.abspath(__file__))
    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)

    meta = real.get_smry_meta(column_keys=["*"])
    assert isinstance(meta, dict)
    assert "FOPT" in meta
    assert "FOPTH" in meta
    assert meta["FOPT"]["unit"] == "SM3"
    assert meta["FOPR"]["unit"] == "SM3/DAY"
    assert meta["FOPT"]["is_total"]
    assert not meta["FOPR"]["is_total"]
    assert not meta["FOPT"]["is_rate"]
    assert meta["FOPR"]["is_rate"]
    assert not meta["FOPT"]["is_historical"]
    assert meta["FOPTH"]["is_historical"]

    assert meta["WOPR:OP_1"]["wgname"] == "OP_1"
    assert meta["WOPR:OP_1"]["keyword"] == "WOPR"
    if "wgname" in meta["FOPT"]:
        # Not enforced yet to have None fields actually included
        assert meta["FOPT"]["wgname"] is None

    # Can create dataframes like this:
    meta_df = pd.DataFrame.from_dict(meta, orient="index")
    hist_keys = meta_df[meta_df["is_historical"]].index
    assert all([key.split(":")[0].endswith("H") for key in hist_keys])

    # When virtualizing a realization, smry data must be loaded
    # for smry metadata to be conserved
    real.load_smry()
    vreal = real.to_virtual()
    vmeta = vreal.get_smry_meta()
    assert "FOPT" in vmeta
    assert vmeta["FOPR"]["unit"] == "SM3/DAY"


def test_get_df_merge():
    """Test that we can merge on the fly using get_df()"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    realdir = os.path.join(testdir, "data/testensemble-reek001", "realization-0/iter-0")
    real = ensemble.ScratchRealization(realdir)
    onlysmry = real.load_smry(column_keys="F*", time_index="monthly")
    assert "parameters.txt" in real.data

    paramcount = len(real.parameters)
    smrycount = len(onlysmry.columns)
    smry = real.get_df("unsmry--monthly", merge="parameters.txt")
    assert len(smry.columns) == paramcount + smrycount
    assert len(smry["SORG1"].unique()) == 1

    # Merge with list should also work:
    smry = real.get_df("unsmry--monthly", merge=["parameters.txt"])
    assert len(smry.columns) == paramcount + smrycount

    # Merge with empty list:
    smry = real.get_df("unsmry--monthly", merge=[])
    assert len(smry.columns) == smrycount

    # Merge with multiple output sets:
    outputs = real.load_txt("outputs.txt")
    smry = real.get_df("unsmry--monthly", merge=["parameters", "outputs"])
    assert len(smry.columns) == paramcount + smrycount + len(outputs)

    # Merge with scalar data, and combination of scalar and dict data:
    real.load_scalar("npv.txt")
    smry = real.get_df("unsmry--monthly", merge="npv.txt")
    assert len(smry.columns) == smrycount + 1
    smry = real.get_df("unsmry--monthly", merge=["parameters.txt", "npv.txt"])
    assert len(smry.columns) == smrycount + paramcount + 1

    # Try merging dataframes:
    real.load_csv("share/results/volumes/simulator_volume_fipnum.csv")

    # Inject a mocked dataframe to the realization:
    real.data["fipnum2zone"] = pd.DataFrame(
        columns=["FIPNUM", "ZONE"],
        data=[
            [1, "UpperReek"],
            [2, "MidReek"],
            [3, "LowerReek"],
            [4, "UpperReek"],
            [5, "MidReek"],
            [6, "LowerReek"],
        ],
    )
    volframe = real.get_df("simulator_volume_fipnum", merge="fipnum2zone")

    assert "ZONE" in volframe
    assert "FIPNUM" in volframe
    assert "STOIIP_OIL" in volframe
    assert len(volframe["ZONE"].unique()) == 3

    scalar_dict = real.get_df("npv.txt", merge="outputs")
    assert "npv.txt" in scalar_dict
    assert "top_structure" in scalar_dict

    # Inject a random dict and merge with:
    real.data["foodict"] = dict(BAR="COM")
    dframe = real.get_df("parameters", merge="foodict")
    assert "BAR" in dframe
    assert "SORG1" in dframe

    # Merge something that is not mergeable
    real.data["randtable"] = pd.DataFrame(
        columns=["BARF", "ARBF"], data=[[1, 3], [2, 4]]
    )
    with pytest.raises(TypeError):
        # pylint: disable=pointless-statement
        real.get_df("parameters", merge="randtable")

    with pytest.raises(pd.errors.MergeError):
        # pylint: disable=pointless-statement
        real.get_df("unsmry--monthly", merge="randtable")



"""This module is deprecated and will be removed in fmu-ensemble v2.0.0"""

import inspect
import logging
import os
import sys
import timeit
import warnings

# pylint: disable=protected-access

warnings.filterwarnings(
    action="always",
    category=DeprecationWarning,
    module=r"etc|ensemble.etc|fmu.ensemble.etc",
)


class _BColors(object):
    # local class for ANSI term color commands
    # bgcolors:
    # 40=black, 41=red, 42=green, 43=yellow, 44=blue, 45=pink, 46 cyan

    # pylint: disable=too-few-public-methods
    HEADER = "\033[1;96m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARN = "\033[93;43m"
    ERROR = "\033[93;41m"
    CRITICAL = "\033[1;91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    def __init__(self):
        pass


class Interaction(object):
    """System for handling interaction; dialogues and messages in FMU.

    This module cooperates with the standard Python logging module.
    """

    def __init__(self):
        warnings.warn(
            "fmu.ensemble.etc is deprecated and will be removed in later versions.",
            DeprecationWarning,
        )
        self._callclass = None
        self._caller = None
        self._lformat = None
        self._lformatlevel = 1
        self._logginglevel = "WARNING"
        self._loggingname = ""
        self._syslevel = 1
        self._test_env = True
        self._tmpdir = "TMP"
        self._testpath = None

        # a string, for Python logging:
        logginglevel = os.environ.get("FMU_LOGGING_LEVEL")

        # a number, for format, 1 is simple, 2 is more info etc
        loggingformat = os.environ.get("FMU_LOGGING_FORMAT")

        if logginglevel is not None:
            self.logginglevel = logginglevel

        if loggingformat is not None:
            self._lformatlevel = int(loggingformat)

    @property
    def logginglevel(self):
        """Set or return a logging level property, e.g. logging.CRITICAL"""

        return self._logginglevel

    @logginglevel.setter
    def logginglevel(self, level):
        # pylint: disable=pointless-statement

        validlevels = ("INFO", "WARNING", "DEBUG", "ERROR", "CRITICAL")
        if level in validlevels:
            self._logginglevel = level
        else:
            raise ValueError(
                "Invalid level given, must be " "in {}".format(validlevels)
            )

    @property
    def numericallogginglevel(self):
        """Return a numerical logging level (read only)"""
        llo = logging.CRITICAL
        if self._logginglevel == "INFO":
            llo = logging.INFO
        elif self._logginglevel == "WARNING":
            llo = logging.WARNING
        elif self._logginglevel == "ERROR":
            llo = logging.ERROR
        elif self._logginglevel == "DEBUG":
            llo = logging.DEBUG

        return llo

    @property
    def loggingformatlevel(self):
        """Set logging format (for future use)"""
        return self._lformatlevel

    @property
    def loggingformat(self):
        """Returns the format string to be used in logging"""

        if self._lformatlevel <= 1:
            self._lformat = "%(levelname)8s: \t%(message)s"
        else:
            self._lformat = (
                "%(asctime)s Line: %(lineno)4d %(name)44s "
                + "[%(funcName)40s()]"
                + "%(levelname)8s:"
                + "\t%(message)s"
            )

        return self._lformat

    @property
    def tmpdir(self):
        """Get and set tmpdir for testing"""
        return self._tmpdir

    @tmpdir.setter
    def tmpdir(self, value):
        self._tmpdir = value

    @staticmethod
    def print_fmu_header(appname, appversion, info=None):
        """Prints a banner for a FMU app to STDOUT.

        Args:
            appname (str): Name of application.
            appversion (str): Version of application on form '3.2.1'
            info (str, optional): More info, e.g. if beta release

        Example::

            fmux.print_fmu_header('fmu.ensemble, '0.2.1', info='Beta release!')
        """
        cur_version = "Python " + str(sys.version_info[0]) + "."
        cur_version += str(sys.version_info[1]) + "." + str(sys.version_info[2])

        app = "This is " + appname + ", v. " + str(appversion)
        if info:
            app = app + " (" + info + ")"

        print("")
        print(_BColors.HEADER)
        print("#" * 79)
        print("#{}#".format(app.center(77)))
        print("#{}#".format(cur_version.center(77)))
        print("#" * 79)
        print(_BColors.ENDC)
        print("")

    def basiclogger(self, name, level=None):
        """Initiate the logger by some default settings."""

        if level is not None:
            self.logginglevel = level

        fmt = self.loggingformat
        self._loggingname = name
        logging.basicConfig(format=fmt, stream=sys.stdout)
        logging.getLogger().setLevel(self.numericallogginglevel)  # root logger
        logging.captureWarnings(True)

        return logging.getLogger(name)

    @staticmethod
    def functionlogger(name):
        """Get the logger for functions (not top level)."""

        logger = logging.getLogger(name)
        logger.addHandler(logging.NullHandler())
        return logger

    def testsetup(self, path="TMP"):
        """Basic setup for FMU testing (developer only; relevant for tests)"""

        try:
            os.makedirs(path)
        except OSError:
            if not os.path.isdir(path):
                raise

        self._test_env = True
        self._tmpdir = path
        self._testpath = None

        return True

    @staticmethod
    def timer(*args):
        """Without args; return the time, with a time as arg return the
        difference.

        Example::

            time1 = timer()
            for i in range(10000):
                i = i + 1
            time2 = timer(time1)
            print('Execution took {} seconds'.format(time2)

        """
        time1 = timeit.default_timer()

        if args:
            return time1 - args[0]

        return time1

    def echo(self, string):
        """Show info at runtime (for user scripts)"""
        level = -5
        idx = 3

        caller = sys._getframe(1).f_code.co_name
        frame = inspect.stack()[1][0]
        self.get_callerinfo(caller, frame)

        self._output(idx, level, string)

    def warn(self, string):
        """Show warnings at Runtime (pure user info/warns)."""
        level = 0
        idx = 6

        caller = sys._getframe(1).f_code.co_name
        frame = inspect.stack()[1][0]
        self.get_callerinfo(caller, frame)

        self._output(idx, level, string)

    warning = warn

    def error(self, string):
        """Issue an error, will not exit system by default"""
        level = -8
        idx = 8

        caller = sys._getframe(1).f_code.co_name
        frame = inspect.stack()[1][0]
        self.get_callerinfo(caller, frame)

        self._output(idx, level, string)

    def critical(self, string, sysexit=True):
        """Issue a critical error, default is SystemExit."""
        level = -9
        idx = 9

        caller = sys._getframe(1).f_code.co_name
        frame = inspect.stack()[1][0]
        self.get_callerinfo(caller, frame)

        self._output(idx, level, string)
        if sysexit:
            raise SystemExit("STOP!")

    def get_callerinfo(self, caller, frame):
        """Get caller info for logging (developer stuff)"""
        the_class = self._get_class_from_frame(frame)

        # just keep the last class element
        xname = str(the_class)
        xname = xname.split(".")
        the_class = xname[-1]

        self._caller = caller
        self._callclass = the_class

        return (self._caller, self._callclass)

    # =========================================================================
    # Private routines
    # =========================================================================

    @staticmethod
    def _get_class_from_frame(frame):
        # Incomplete (need more coffee)
        current = inspect.currentframe()
        outer = inspect.getouterframes(current)
        return outer[0]

    def _output(self, idx, level, string):
        # pylint: disable=too-many-branches

        prefix = ""
        endfix = ""

        if idx == 0:
            prefix = "++"
        elif idx == 1:
            prefix = "**"
        elif idx == 3:
            prefix = ">>"
        elif idx == 6:
            prefix = _BColors.WARN + "##"
            endfix = _BColors.ENDC
        elif idx == 8:
            prefix = _BColors.ERROR + "!#"
            endfix = _BColors.ENDC
        elif idx == 9:
            prefix = _BColors.CRITICAL + "!!"
            endfix = _BColors.ENDC

        prompt = False
        if level <= self._syslevel:
            prompt = True

        if prompt:
            if self._syslevel <= 1:
                print("{} {}{}".format(prefix, string, endfix))
            else:
                ulevel = str(level)
                if level == -5:
                    ulevel = "M"
                if level == -8:
                    ulevel = "E"
                if level == -9:
                    ulevel = "W"
                print(
                    "{0} <{1}> [{2:23s}->{3:>33s}] {4}{5}".format(
                        prefix, ulevel, self._callclass, self._caller, string, endfix
                    )
                )



jobList = [ {"name" : "DESIGN2PARAMS",
  "executable" : "/project/res/etc/ERT/Scripts/design2params.py",
  "target_file" : "DESIGN2PARAMS.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN2PARAMS.stdout",
  "stderr" : "DESIGN2PARAMS.stderr",
  "stdin" : None,
  "argList" : ["2","/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/distributions/design_matrix.xlsx","DesignMatrix","DefaultValues"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "DESIGN_KW",
  "executable" : "/project/res/etc/ERT/Scripts/design_kw.py",
  "target_file" : "DESIGN_KW.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN_KW.stdout",
  "stderr" : "DESIGN_KW.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/templates/global_variables.tmpl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/global_variables.ipl"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "DESIGN_KW",
  "executable" : "/project/res/etc/ERT/Scripts/design_kw.py",
  "target_file" : "DESIGN_KW.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN_KW.stdout",
  "stderr" : "DESIGN_KW.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/templates/multflt.tmpl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/reek.multflt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "DESIGN_KW",
  "executable" : "/project/res/etc/ERT/Scripts/design_kw.py",
  "target_file" : "DESIGN_KW.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN_KW.stdout",
  "stderr" : "DESIGN_KW.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/templates/let-swof.tmpl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/let-swof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "DESIGN_KW",
  "executable" : "/project/res/etc/ERT/Scripts/design_kw.py",
  "target_file" : "DESIGN_KW.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN_KW.stdout",
  "stderr" : "DESIGN_KW.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/templates/let-sgof.tmpl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/let-sgof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "DESIGN_KW",
  "executable" : "/project/res/etc/ERT/Scripts/design_kw.py",
  "target_file" : "DESIGN_KW.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN_KW.stdout",
  "stderr" : "DESIGN_KW.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/templates/multz.tmpl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/reek.multz"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/model"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_directory.stdout",
  "stderr" : "copy_directory.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../rms/bin","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/rms"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_directory.stdout",
  "stderr" : "copy_directory.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../rms/ipl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/rms"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/input/global_variables"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_directory.stdout",
  "stderr" : "copy_directory.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../rms/input/config","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/rms/input"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_directory.stdout",
  "stderr" : "copy_directory.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../rms/input/well_modelling","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/rms/input"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/aps"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/cohiba/surfaces"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/cohiba/logfiles"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/pem"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/regions/data/grid"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/regions/result"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["rms/output/zone"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["sim2seis/model"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_file.stdout",
  "stderr" : "copy_file.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../sim2seis/model/model_2000_sim.txt","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/sim2seis/model/model_2000_sim.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_file.stdout",
  "stderr" : "copy_file.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../sim2seis/model/model_2001_sim.txt","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/sim2seis/model/model_2001_sim.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_file.stdout",
  "stderr" : "copy_file.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../sim2seis/model/model_2003_sim.txt","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/sim2seis/model/model_2003_sim.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["sim2seis/input"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["sim2seis/output"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["eclipse/include/grid"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["eclipse/include/props"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_file.stdout",
  "stderr" : "copy_file.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../eclipse/include/props/reek.endpoints","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/reek.endpoints"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_file.stdout",
  "stderr" : "copy_file.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../eclipse/include/props/reek.pvt","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/reek.pvt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["eclipse/include/regions"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["eclipse/include/solution"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "COPY_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/copy_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "copy_directory.stdout",
  "stderr" : "copy_directory.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../eclipse/include/summary","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["eclipse/include/schedule"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["share/results/maps"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["share/results/grids"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["share/results/polygons"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["share/results/volumes"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["share/results/images"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MAKE_DIRECTORY",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/make_directory.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "make_directory.stdout",
  "stderr" : "make_directory.stderr",
  "stdin" : None,
  "argList" : ["share/results/observations"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MOVE_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/move_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "move_file.stdout",
  "stderr" : "move_file.stderr",
  "stdin" : None,
  "argList" : ["global_variables.ipl","rms/input/global_variables/global_variables.ipl"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MOVE_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/move_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "move_file.stdout",
  "stderr" : "move_file.stderr",
  "stdin" : None,
  "argList" : ["reek.multflt","eclipse/include/grid/reek.multflt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MOVE_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/move_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "move_file.stdout",
  "stderr" : "move_file.stderr",
  "stdin" : None,
  "argList" : ["let-swof.txt","eclipse/include/props/let-swof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MOVE_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/move_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "move_file.stdout",
  "stderr" : "move_file.stderr",
  "stdin" : None,
  "argList" : ["let-sgof.txt","eclipse/include/props/let-sgof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "MOVE_FILE",
  "executable" : "/project/res/etc/ERT/Config/jobs/util/script/move_file.py",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "move_file.stdout",
  "stderr" : "move_file.stderr",
  "stdin" : None,
  "argList" : ["reek.multz","eclipse/include/grid/reek.multz"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "DESIGN_KW",
  "executable" : "/project/res/etc/ERT/Scripts/design_kw.py",
  "target_file" : "DESIGN_KW.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "DESIGN_KW.stdout",
  "stderr" : "DESIGN_KW.stderr",
  "stdin" : None,
  "argList" : ["/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../input/templates/seed.tmpl","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/rms/model/RMS_SEED"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "RMS_BATCH",
  "executable" : "/project/res/etc/ERT/Scripts/run_rms_batch.py",
  "target_file" : "RMS_TARGET.INC",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "rms.stdout",
  "stderr" : "rms.stderr",
  "stdin" : None,
  "argList" : ["rms/model","2","10.1.1","/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/model/../../rms/model/reek.rms10.1.1","./","./","MAIN"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "GENERATE_RELPERM",
  "executable" : "/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/bin/scripts/generate_relperm.sh",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "GENERATE_RELPERM.stdout",
  "stderr" : "GENERATE_RELPERM.stderr",
  "stdin" : None,
  "argList" : ["/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/let-swof.txt","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/swof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "GENERATE_RELPERM",
  "executable" : "/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/bin/scripts/generate_relperm.sh",
  "target_file" : None,
  "error_file" : None,
  "start_file" : None,
  "stdout" : "GENERATE_RELPERM.stdout",
  "stderr" : "GENERATE_RELPERM.stderr",
  "stdin" : None,
  "argList" : ["/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/let-sgof.txt","/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/sgof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "INCLUDE_PC",
  "executable" : "/project/fmu/users/havb/reek_20180207/resmod/ff/2017a/r001/ert/bin/scripts/include_pc.py",
  "target_file" : "INCLUDE_PC.INC",
  "error_file" : None,
  "start_file" : None,
  "stdout" : "INCLUDE_PC.stdout",
  "stderr" : "INCLUDE_PC.stderr",
  "stdin" : None,
  "argList" : ["/scratch/fmu/havb/2_r001_reek/realization-2/iter-0/eclipse/include/props/swof.txt"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
},
 {"name" : "ECLIPSE100_2014.2",
  "executable" : "/project/res/etc/ERT/Config/jobs/ressim/Scripts/run_ecl100",
  "target_file" : "eclipse/model/2_R001_REEK-2.OK",
  "error_file" : None,
  "start_file" : None,
  "stdout" : None,
  "stderr" : "ECLIPSE.stderr",
  "stdin" : None,
  "argList" : ["2014.2","eclipse/model/2_R001_REEK-2","1"],
  "environment" : None,
  "license_path" : None,
  "max_running_minutes" : None,
  "max_running" : None,
}]
umask = 0000



"""Module for handling linear combinations of ensembles"""

import fnmatch
import logging

import pandas as pd

logger = logging.getLogger(__name__)


class EnsembleCombination(object):
    """The class is used to perform linear operations on ensembles.

    When instantiated, the linear combination will not actually be
    computed before the results are actually asked for - lazy
    evaluation.
    """

    def __init__(self, ref, scale=None, add=None, sub=None):
        """Set up an object for a linear combination of ensembles.

        Each instance of this object can only hold one operation,
        either addition/substraction of two
        ensembles/ensemblecombinations or a scaling of one.

        ScratchEnsembles and VirtualEnsembles can be combined freely.

        A long expression of ensembles will lead to an evaluation tree
        consisting of instances of this class with actual ensembles
        at the leaf nodes.

        Args:
            scale: float for scaling the ensemble or ensemblecombination
            add: ensemble or ensemblecombinaton with a positive sign
            sub: ensemble or ensemblecombination with a negative sign.

        """
        self.ref = ref
        if scale:
            self.scale = scale
        else:
            self.scale = 1

        if add:
            self.add = add
        else:
            self.add = None

        # Alternatively, substraction could be implemented as a combination
        # of __mult__ and __add__
        if sub:
            self.sub = sub
        else:
            self.sub = None

    def keys(self):
        """Return the intersection of all keys available in reference
        ensemble(combination) and the other
        """
        combkeys = set()
        combkeys = combkeys.union(self.ref.keys())
        if self.add:
            combkeys = combkeys.intersection(self.add.keys())
        if self.sub:
            combkeys = combkeys.intersection(self.sub.keys())
        return combkeys

    def get_df(self, localpath, merge=None):
        """Obtain given data from the ensemblecombination,
        doing the actual computation of ensemble on the fly.

        Warning: In order to add dataframes together with meaning,
        using pandas.add, the index of the frames must be correctly set,
        and this can be tricky for some datatypes (f.ex. volumetrics table
        where you want to add together volumes for correct zone
        and fault segment).

        If you have the columns "REAL", "DATE", "ZONE" and/or "REGION", it
        will be regarded as an index column.

        Args:
            localpath (str): refers to the internalized name of the
                data wanted in each ensemble.
            merge (list or str): Optional data to be merged in for the data
                The merge will happen as deep as possible (in realization
                objects in case of ScratchEnsembles), and all ensemble
                combination computations happen after merging. Be careful
                with index guessing and merged data.
        """
        # We can pandas.add when the index is set correct.
        # WE MUST GUESS!
        indexlist = []
        indexcandidates = ["REAL", "DATE", "ZONE", "REGION"]
        for index in indexcandidates:
            if index in self.ref.get_df(localpath).columns:
                indexlist.append(index)
        logger.debug("get_df() inferred index columns to %s", str(indexlist))
        refdf = self.ref.get_df(localpath, merge=merge).set_index(indexlist)
        refdf = refdf.select_dtypes(include="number")
        result = refdf.mul(self.scale)
        if self.add:
            otherdf = self.add.get_df(localpath, merge=merge).set_index(indexlist)
            otherdf = otherdf.select_dtypes(include="number")
            result = result.add(otherdf)
        if self.sub:
            otherdf = self.sub.get_df(localpath, merge=merge).set_index(indexlist)
            otherdf = otherdf.select_dtypes(include="number")
            result = result.sub(otherdf)
        # Delete rows where everything is NaN, which will be case when
        # realization (multi-)indices does not match up in both ensembles.
        result.dropna(axis="index", how="all", inplace=True)
        # Also delete columns where everything is NaN, happens when
        # column data are not similar
        result.dropna(axis="columns", how="all", inplace=True)
        return result.reset_index()

    def to_virtual(self, keyfilter=None):
        """Evaluate the current linear combination and return as
        a virtual ensemble.

        Args:
            keyfilter (list or str): If supplied, only keys matching wildcards
                in this argument will be included. Use this for speed reasons
                when only some data is needed. Default is to include everything.
                If you supply "unsmry", it will match every key that
                includes this string by prepending and appending '*' to your pattern

        Returns:
            VirtualEnsemble
        """
        # pylint: disable=import-outside-toplevel
        from .virtualensemble import VirtualEnsemble

        if keyfilter is None:
            keyfilter = "*"
        if isinstance(keyfilter, str):
            keyfilter = [keyfilter]
        if not isinstance(keyfilter, list):
            raise TypeError("keyfilter in to_virtual() must be list or string")

        vens = VirtualEnsemble(name=str(self))
        for key in self.keys():
            if sum(
                [fnmatch.fnmatch(key, "*" + pattern + "*") for pattern in keyfilter]
            ):
                logger.info("Calculating ensemblecombination on %s", key)
                vens.append(key, self.get_df(key))
        vens.update_realindices()
        return vens

    def get_smry_dates(
        self, freq="monthly", normalize=True, start_date=None, end_date=None
    ):
        """Create a union of dates available in the
        involved ensembles
        """
        dates = set(self.ref.get_smry_dates(freq, normalize, start_date, end_date))
        if self.add:
            dates = dates.union(
                set(self.add.get_smry_dates(freq, normalize, start_date, end_date))
            )
        if self.sub:
            dates = dates.union(
                set(self.sub.get_smry_dates(freq, normalize, start_date, end_date))
            )
        dates = list(dates)
        dates.sort()
        return dates

    def get_smry(self, column_keys=None, time_index=None):
        """
        Loads the Eclipse summary data directly from the underlying
        ensemble data. The ensembles can be ScratchEnsemble or
        VirtualEnsemble, if scratch it will access binary
        summary files directly, if virtual ensembles, summary
        data must have been loaded earlier.

        Args:
            column_keys (str or list): column key wildcards. Default
                is '*', which will match all vectors in the Eclipse
                output.
            time_index (str or list of DateTime): time_index mnemonic or
                a list of explicit datetime at which the summary data
                is requested (interpolated or extrapolated)

        Returns:
            pd.DataFrame. Indexed by rows, has at least the columns REAL
                and DATE if not empty.
        """
        if isinstance(time_index, str):
            time_index = self.get_smry_dates(time_index)
        indexlist = ["REAL", "DATE"]
        refdf = self.ref.get_smry(
            time_index=time_index, column_keys=column_keys
        ).set_index(indexlist)
        result = refdf.mul(self.scale)
        if self.add:
            otherdf = self.add.get_smry(
                time_index=time_index, column_keys=column_keys
            ).set_index(indexlist)
            result = result.add(otherdf)
        if self.sub:
            otherdf = self.sub.get_smry(
                time_index=time_index, column_keys=column_keys
            ).set_index(indexlist)
            result = result.sub(otherdf)
        return result.reset_index()

    def get_smry_stats(self, column_keys=None, time_index="monthly"):
        """
        Function to extract the ensemble statistics (Mean, Min, Max, P10, P90)
        for a set of simulation summary vectors (column key).

        Compared to the agg() function, this function only works on summary
        data (time series), and will only operate on actually requested data,
        independent of what is internalized. It accesses the summary files
        directly and can thus obtain data at any time frequency.

        Args:
            column_keys: list of column key wildcards
            time_index: list of DateTime if interpolation is wanted
               default is None, which returns the raw Eclipse report times
               If a string is supplied, that string is attempted used
               via get_smry_dates() in order to obtain a time index.
        Returns:
            A MultiLevel dataframe. Outer index is 'minimum', 'maximum',
            'mean', 'p10', 'p90', inner index are the dates. Column names
            are the different vectors. Quantiles follow the scientific
            standard, opposite to the oil industry standard.

        TODO: add warning message when failed realizations are removed
        """
        # Obtain an aggregated dataframe for only the needed columns over
        # the entire ensemble.

        dframe = (
            self.get_smry(time_index=time_index, column_keys=column_keys)
            .drop(columns="REAL")
            .groupby("DATE")
        )
        mean = dframe.mean()
        p90 = dframe.quantile(q=0.90)
        p10 = dframe.quantile(q=0.10)
        maximum = dframe.max()
        minimum = dframe.min()

        return pd.concat(
            [mean, p10, p90, maximum, minimum],
            keys=["mean", "p10", "p90", "maximum", "minimum"],
            names=["statistic"],
            sort=False,
        )

    def get_smry_meta(self, column_keys=None):
        """
        Provide metadata for summary data vectors.

        A dictionary indexed by summary vector names is returned, and each
        value is another dictionary with potentially the metadata types:
        * unit (string)
        * is_total (bool)
        * is_rate (bool)
        * is_historical (bool)
        * get_num (int) (only provided if not None)
        * keyword (str)
        * wgname (str or None)

        Args:
            column_keys: List or str of column key wildcards
        """
        meta = self.ref.get_smry_meta(column_keys=column_keys)
        if self.add:
            meta.update(self.add.get_smry_meta(column_keys=column_keys))
        if self.sub:
            meta.update(self.sub.get_smry_meta(column_keys=column_keys))
        return meta

    def agg(self, aggregation, keylist=None, excludekeys=None):
        """Aggregator, this is a wrapper that will
        call .to_virtual() on your behalf and call the corresponding
        agg() in VirtualEnsemble.
        """
        return self.to_virtual().agg(aggregation, keylist, excludekeys)

    def get_volumetric_rates(
        self, column_keys=None, time_index="monthly", time_unit=None
    ):
        """Compute volumetric rates from cumulative summary
        vectors.

        Column names that are not referring to cumulative summary
        vectors are silently ignored.

        A Dataframe is returned with volumetric rates, that is rate
        values that can be summed up to the cumulative version. The
        'T' in the column name is switched with 'R'. If you ask for
        FOPT, you will get FOPR in the returned dataframe.

        Rates in the returned dataframe are valid **forwards** in time,
        opposed to rates coming directly from the Eclipse simulator which
        are valid backwards in time.

        If time_unit is set, the rates will be scaled to represent
        either daily, monthly or yearly rates. These will sum up to the
        cumulative as long as you multiply with the correct number
        of days, months or year between each consecutive date index.
        Month lengths and leap years are correctly handled.

        Args:
            column_keys: str or list of strings, cumulative summary vectors
            time_index: str or list of datetimes
            time_unit: str or None. If None, the rates returned will
                be the difference in cumulative between each included
                time step (where the time interval can vary arbitrarily)
                If set to 'days', 'months' or 'years', the rates will
                be scaled to represent a daily, monthly or yearly rate that
                is compatible with the date index and the cumulative data.

        """
        return self.to_virtual(keyfilter="unsmry").get_volumetric_rates(
            column_keys=column_keys, time_index=time_index, time_unit=time_unit
        )

    @property
    def parameters(self):
        """Return parameters from the ensemble as a class property"""
        try:
            return self.get_df("parameters.txt")
        except KeyError:
            return pd.DataFrame()

    def __len__(self):
        """Estimate the number of realizations in this
        ensemble combinations.

        This is not always well defined in cases of strange
        combinations of which data is available in which realization,
        so after actual computation of a virtual ensemble, the number
        of realizations can be less that what this estimate returns

        Returns:
            int, number of realizations (upper limit)
        """
        return len(self.get_realindices())

    def get_realindices(self):
        """Return the integer indices for realizations in this ensemble

        There is no guarantee that all realizations returned here
        will be valid for all datatypes after computation.

        Returns:
            list of integers
        """
        indices = set(self.ref.get_realindices())
        if self.add:
            indices = indices.intersection(set(self.add.get_realindices()))
        if self.sub:
            indices = indices.intersection(set(self.sub.get_realindices()))
        return list(indices)

    def __getitem__(self, localpath):
        return self.get_df(localpath)

    def __repr__(self):
        """Try to give out a linear expression"""
        # NB: Implementation in this method requires scaling not to happen
        # simultaneously as adds or subs.
        scalestring = ""
        addstring = ""
        substring = ""
        if self.scale != 1:
            scalestring = str(self.scale) + " * "
        if self.add:
            addstring = " + " + str(self.add)
        if self.sub:
            substring = " - " + str(self.sub)
        return scalestring + str(self.ref) + addstring + substring

    def __sub__(self, other):
        """Substract another ensemble from this combination"""
        return EnsembleCombination(self, sub=other)

    def __add__(self, other):
        """Add another ensemble from this combination"""
        return EnsembleCombination(self, add=other)

    def __radd__(self, other):
        """Add another ensemble from this combination"""
        return EnsembleCombination(self, add=other)

    def __rsub__(self, other):
        """Substract another ensemble from this combination"""
        return EnsembleCombination(self, sub=other)

    def __mul__(self, other):
        """Scale this EnsembleCombination by a scalar value"""
        return EnsembleCombination(self, scale=float(other))

    def __rmul__(self, other):
        """Scale this EnsembleCombination by a scalar value"""
        return EnsembleCombination(self, scale=float(other))



"""Test a deprecated submodule"""

import pytest
from fmu import ensemble


def test_deprecated_etc():
    """The following lines was in documentation up to 1.4.0 inclusive.

    v1.4.1 and v1.4.2 miss support for fmu.ensemble.etc.Interaction()
    and then it is reinstated in v1.4.3. To be removed for fmu.ensemble v2.0.0
    """

    with pytest.warns(DeprecationWarning):
        fmux = ensemble.etc.Interaction()
    logger = fmux.basiclogger(__name__, level="WARNING")
    logger.info("testing deprecated code")

    # pylint: disable=import-outside-toplevel
    # Test different import:
    from fmu.ensemble import etc

    with pytest.warns(DeprecationWarning):
        fmux = etc.Interaction()
    logger = fmux.basiclogger(__name__, level="WARNING")
    logger.info("testing deprecated code")



"""Testing fmu-ensemble."""

import logging
import os

import pytest
from fmu import ensemble
from fmu.ensemble import ScratchEnsemble

logger = logging.getLogger(__name__)


def test_realizationcombination_basic():
    """Basic testing of combination of two realizations
    to a RealizationCombination"""

    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    real0dir = os.path.join(
        testdir, "data/testensemble-reek001", "realization-0/iter-0"
    )
    real0 = ensemble.ScratchRealization(real0dir)
    real0.load_smry(time_index="yearly", column_keys=["F*"])
    real0.load_scalar("npv.txt")
    real1dir = os.path.join(
        testdir, "data/testensemble-reek001", "realization-1/iter-0"
    )
    real1 = ensemble.ScratchRealization(real1dir)
    real1.load_smry(time_index="yearly", column_keys=["F*"])
    real1.load_scalar("npv.txt")
    realdiff = real0 - real1
    real0.data["a_string"] = "foo_in_real_0"
    real1.data["a_string"] = "foo_in_real_1"
    assert "FWPR" in realdiff["unsmry--yearly"]
    assert "FWL" in realdiff["parameters"]

    assert realdiff["npv.txt"] == real0["npv.txt"] - real1["npv.txt"]
    assert realdiff["a_string"] is None
    # Combination of the same when virtualized:
    vreal0 = real0.to_virtual()
    vreal1 = real1.to_virtual()

    scaled_vreal0 = 3 * vreal0
    assert "FWPR" in scaled_vreal0["unsmry--yearly"]
    assert "FWL" in scaled_vreal0["parameters"]
    assert "FWL" in scaled_vreal0.parameters
    assert scaled_vreal0.parameters["FWL"] == real0.parameters["FWL"] * 3

    vdiff = vreal1 - vreal0
    assert "FWPR" in vdiff["unsmry--yearly"]
    assert "FWL" in vdiff["parameters"]
    assert vdiff["npv.txt"] == real1["npv.txt"] - real0["npv.txt"]
    vdiff_filtered = vdiff.to_virtual(keyfilter="parameters")
    assert "parameters.txt" in vdiff_filtered.keys()
    with pytest.raises((KeyError, ValueError)):
        vdiff_filtered.get_df("unsmry--yearly")

    vdiff_filtered2 = vdiff.to_virtual(keyfilter="unsmry--yearly")
    assert "parameters.txt" not in vdiff_filtered2.keys()
    assert "FWPR" in vdiff_filtered2.get_df("unsmry--yearly")

    smrymeta = realdiff.get_smry_meta(["FO*"])
    assert "FOPT" in smrymeta

    smry_params = realdiff.get_df("unsmry--yearly", merge="parameters.txt")
    assert "SORG1" in smry_params
    assert "FWCT" in smry_params


def test_realizationcomb_virt_meta():
    """Test metadata aggregation of combinations
    of virtualized realizations"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    real0dir = os.path.join(
        testdir, "data/testensemble-reek001", "realization-0/iter-0"
    )
    real0 = ensemble.ScratchRealization(real0dir)
    real0.load_smry(time_index="yearly", column_keys=["F*"])
    real1dir = os.path.join(
        testdir, "data/testensemble-reek001", "realization-1/iter-0"
    )
    real1 = ensemble.ScratchRealization(real1dir)
    real1.load_smry(time_index="yearly", column_keys=["FOPT", "WOPT*"])

    # Virtualized based on the loades summary vectors, which
    # differ between the two realizations.
    vreal0 = real0.to_virtual()
    vreal1 = real1.to_virtual()

    assert "WOPT" not in vreal0.get_smry_meta(column_keys="*")
    assert "FOPT" in vreal0.get_smry_meta(column_keys="*")
    assert "WOPT:OP_3" in vreal1.get_smry_meta(column_keys="*")
    assert "WOPT:OP_3" not in vreal0.get_smry_meta(column_keys="*")
    assert "FOPT" in vreal1.get_smry_meta(column_keys="*")


def test_manual_aggregation():
    """Test that aggregating an ensemble using
    RealizationCombination is the same as calling agg() on the
    ensemble"""
    if "__file__" in globals():
        # Easen up copying test code into interactive sessions
        testdir = os.path.dirname(os.path.abspath(__file__))
    else:
        testdir = os.path.abspath(".")

    reekensemble = ScratchEnsemble(
        "reektest", testdir + "/data/testensemble-reek001/" + "realization-*/iter-0"
    )
    reekensemble.load_smry(time_index="yearly", column_keys=["F*"])
    reekensemble.load_csv("share/results/volumes/simulator_volume_fipnum.csv")

    # Aggregate an ensemble into a virtual "mean" realization
    mean = reekensemble.agg("mean")

    # Combine the ensemble members directly into a mean computation.
    # Also returns a virtual realization.
    manualmean = (
        1
        / 5
        * (
            reekensemble[0]
            + reekensemble[1]
            + reekensemble[2]
            + reekensemble[3]
            + reekensemble[4]
        )
    )

    # Commutativity proof:
    assert mean["parameters"]["RMS_SEED"] == manualmean["parameters"]["RMS_SEED"]



"""Contains the VirtualRealization class"""

import fnmatch
import logging
import os
import shutil
import warnings

import numpy as np
import pandas as pd

from .realizationcombination import RealizationCombination
from .util import shortcut2path
from .util.dates import date_range
from .util.rates import compute_volumetric_rates

logger = logging.getLogger(__name__)


class VirtualRealization(object):
    """A computed or archived realization.

    Computed or archived, one cannot assume to have access to the file
    system containing original data.

    Datatables that in a ScratchRealization was available through the
    files dataframe, is now available as dataframes in a dict accessed
    by the localpath in the files dataframe from ScratchRealization-

    """

    def __init__(self, description=None, data=None, longdescription=None):
        self._description = description
        self._longdescription = longdescription
        if data:
            self.data = data
        else:
            self.data = {}

    def keys(self):
        """Return the keys of all data in internal datastore"""
        return self.data.keys()

    def __getitem__(self, localpath):
        """Retrieve data for a specific key. Wrapper for get_df(),
        shorthands are allowed."""
        return self.get_df(localpath)

    def __delitem__(self, localpath):
        """Delete a key from the internal datastore. The key must be fully
        qualified, no shorthands."""
        if localpath in self.keys():
            del self.data[localpath]

    def append(self, key, dataframe, overwrite=False):
        """Append data to the datastore.

        No checks performed on the dataframe coming in. If key exists,
        nothing will be appended unless overwrite is set to True
        """
        if key in self.data.keys() and not overwrite:
            logger.warning("Ignoring %s, data already exists", key)
            return
        self.data[key] = dataframe

    def __repr__(self):
        """Represent the realization. Show only the last part of the path"""
        return "<VirtualRealization, {}>".format(self._description)

    def to_disk(self, filesystempath, delete=False):
        """Write the virtual realization to the filesystem.

        All data will be dumped to the requested directory according
        to their localpaths (keys).

        Args:
            filesystempath : string with a directory, absolute or
                relative. If it exists already, it must be empty,
                otherwise we give up.
            delete: boolean, if True, existing directory at the filesystempath
                will be deleted before export.
        """
        warnings.warn(
            "to_disk() is considered deprecated and might be "
            "removed in fmu-ensemble v2.0.0",
            FutureWarning,
        )
        if os.path.exists(filesystempath):
            if delete:
                shutil.rmtree(filesystempath)
                os.mkdir(filesystempath)
            else:
                if os.listdir(filesystempath):
                    logger.critical("Refusing to write to non-empty directory")
                    raise IOError("Directory %s not empty" % filesystempath)
        else:
            os.mkdir(filesystempath)

        if self._description is not None:
            with open(os.path.join(filesystempath, "_description"), "w") as fhandle:
                fhandle.write(self._description)
        if self._longdescription is not None:
            with open(os.path.join(filesystempath, "_longdescription"), "w") as fhandle:
                fhandle.write(str(self._longdescription))
        with open(os.path.join(filesystempath, "__repr__"), "w") as fhandle:
            fhandle.write(self.__repr__())

        for key in self.keys():
            dirname = os.path.join(filesystempath, os.path.dirname(key))
            if dirname:
                if not os.path.exists(dirname):
                    os.makedirs(dirname)

            data = self.get_df(key)
            filename = os.path.join(dirname, os.path.basename(key))
            if isinstance(data, pd.DataFrame):
                logger.info("Dumping %s", key)
                data.to_csv(filename, index=False)
            elif isinstance(data, dict):
                with open(filename, "w") as fhandle:
                    for paramkey in data.keys():
                        fhandle.write(paramkey + " " + str(data[paramkey]) + "\n")
            elif isinstance(data, (str, float, int, np.number)):
                with open(filename, "w") as fhandle:
                    fhandle.write(str(data))
            else:
                logger.warning(
                    "Don't know how to dump %s of type %s to disk", key, type(key)
                )

    def load_disk(self, filesystempath):
        """Load data for a virtual realization from disk.

        Existing data in the current object will be wiped,
        this function is intended for initialization

        WARNING: This code is really shaky. We need metafiles written
        by to_json() for robust parsing of files on disk, f.ex. are
        txt files really key-value data (dicts) or csv files?

        Currently, the file format is guessed based on the contents
        of the two first lines:
        * CSV files contains commas, and more than one line
        * key-value files contains two space-separated values, and at least one line
        * scalar files contain only one item and one line

        Args:
            filesystempath: path to a directory that to_disk() has
                written to (or a really careful user)
        """
        warnings.warn(
            "load_disk() is considered deprecated and might be "
            "removed in fmu-ensemble v2.0.0",
            FutureWarning,
        )
        logger.info("Loading virtual realization from %s", filesystempath)
        for root, _, filenames in os.walk(filesystempath):
            for filename in filenames:
                if filename == "_description":
                    self._description = " ".join(
                        open(os.path.join(root, filename)).readlines()
                    )
                    logger.info("got name as %s", self._description)
                elif filename == "STATUS":
                    self.append("STATUS", pd.read_csv(os.path.join(root, filename)))
                    logger.info("got STATUS")
                elif filename == "__repr__":
                    # Not implemented..
                    continue
                else:
                    # GUESS scalar, key-value txt or CSV from the first
                    # two lines. SHAKY!
                    with open(os.path.join(root, filename)) as realfile:
                        lines = realfile.readlines()

                    linecount = len(lines)
                    commafields = len(lines[0].split(","))
                    spacefields = len(lines[0].split())

                    print(filename, commafields, spacefields, linecount)
                    if spacefields == 2 and commafields == 1:
                        # key-value txt file!
                        self.append(
                            filename,
                            pd.read_csv(
                                os.path.join(root, filename),
                                sep=r"\s+",
                                index_col=0,
                                header=None,
                            )[1].to_dict(),
                        )
                        logger.info("Read txt file %s", filename)
                    elif spacefields == 1 and linecount == 1 and commafields == 1:
                        # scalar file
                        value = pd.read_csv(
                            os.path.join(root, filename),
                            sep=r"\s+",
                            header=None,
                            engine="python",
                        ).iloc[0, 0]
                        logger.info("Read scalar file %s", filename)
                        self.append(filename, value)
                    elif spacefields == 1 and linecount > 1 and commafields > 1:
                        # CSV file!
                        self.append(filename, pd.read_csv(os.path.join(root, filename)))
                        logger.info("Read csv file %s", filename)

    def to_json(self):
        """
        Dump realization data to json.

        Resulting json string is compatible with the
        accompanying load_json() function
        """
        raise NotImplementedError

    def get_df(self, localpath, merge=None):
        """Access the internal datastore which contains dataframes, dicts
        or scalars.

        The localpath argument can be shortened, as it will be
        looked up using the function shortcut2path()

        Args:
            localpath (str): the identifier of the data requested
            merge (list or str): refer to additional localpath(s) which
                will be merged into the dataframe.

        Returns:
            dataframe, dictionary, float, str or integer

        Raises:
            KeyError if data is not found.
        """
        data = None
        fullpath = shortcut2path(self.keys(), localpath)
        if fullpath in self.keys():
            data = self.data[fullpath]
        else:
            raise KeyError("Could not find {}".format(localpath))
        if merge is None:
            merge = []
        if not isinstance(merge, list):
            merge = [merge]  # can still be None
        if merge and merge[0] is not None:
            # Make a copy of the primary data to ensure we don't edit it during merge
            if isinstance(data, (pd.DataFrame, dict)):
                data = data.copy()
            elif isinstance(data, (str, int, float, np.number)):
                # Convert scalar data into something mergeable
                value = data
                data = {localpath: value}
            else:
                raise TypeError
        for mergekey in merge:
            if mergekey is None:
                continue
            mergedata = self.get_df(mergekey)
            if isinstance(mergedata, dict):
                for key in mergedata:
                    data[key] = mergedata[key]
            elif isinstance(mergedata, (str, int, float, np.number)):
                # Scalar data, use the mergekey as column
                data[mergekey] = mergedata
            elif isinstance(mergedata, pd.DataFrame):
                data = pd.merge(data, mergedata)
            else:
                raise ValueError("Don't know how to merge data {}".format(mergekey))
        if isinstance(data, pd.Series):
            data = data.to_dict()
        if data is not None:
            return data
        raise ValueError("Data ended up being None")

    def get_volumetric_rates(self, column_keys=None, time_index=None, time_unit=None):
        """Compute volumetric rates from cumulative summary vectors

        See :meth:`fmu.ensemble.util.compute_volumetric_rates`
        """
        return compute_volumetric_rates(self, column_keys, time_index, time_unit)

    def get_smry(self, column_keys=None, time_index="monthly"):
        """Analog function to get_smry() in ScratchRealization

        Accesses the internalized summary data and performs
        interpolation if needed.

        Returns data for those columns that are known, unknown
        columns will be issued a warning for.

        BUG: If some columns are available only in certain dataframes,
        we might miss them (e.g. we ask for yearly FOPT, and we have
        yearly smry with only WOPT data, and FOPT is only in daily
        smry). Resolution is perhaps to merge all relevant data
        upfront.

        Args:
            column_keys: str or list of str with column names,
                may contain wildcards (glob-style). Default is
                to match every key that is known (contrary to
                behaviour in a ScratchRealization)
            time_index: str or list of datetimes

        """
        if not column_keys:
            column_keys = "*"  # Match everything

        column_keys = self._glob_smry_keys(column_keys)
        if not column_keys:
            raise ValueError("No column keys found")

        # If time_index is None, load_smry in ScratchRealization stores as "raw"
        if time_index is None:
            time_index = "raw"

        if isinstance(time_index, str):
            time_index_dt = self.get_smry_dates(time_index)
        elif isinstance(time_index, (list, np.ndarray)):
            time_index_dt = time_index
        else:
            raise TypeError

        # Determine which of the internalized dataframes we should use
        # for interpolation. Or, should we merge some of them for even
        # higher accuracy?

        # Get a list ala ['yearly', 'daily']
        available_smry = [
            x.split("/")[-1].replace(".csv", "").replace("unsmry--", "")
            for x in self.keys()
            if "unsmry" in x
        ]

        if (
            isinstance(time_index, str) and time_index not in available_smry
        ) or isinstance(time_index, list):
            # Suboptimal code, we always pick the finest available
            # time resolution:
            priorities = ["raw", "daily", "weekly", "monthly", "yearly", "custom"]
            # (could also sort them by number of rows, or we could
            #  even merge them all)
            # (could have priorities as a dict, for example so we
            #  can interpolate from monthly if we ask for yearly)
            chosen_smry = ""
            for candidate in priorities:
                if candidate in available_smry:
                    chosen_smry = candidate
                    break
            if not chosen_smry:
                logger.error("No internalized summary data to interpolate from")
                return pd.DataFrame()
        else:
            chosen_smry = time_index

        logger.info(
            "Using %s for interpolation of %s in REAL %s",
            chosen_smry,
            time_index,
            str(self),
        )

        smry = self.get_df("unsmry--" + chosen_smry)[["DATE"] + column_keys]

        # Add the extra datetimes to interpolate at.
        smry.set_index("DATE", inplace=True)
        smry.index = pd.to_datetime(smry.index)
        smry = pd.concat(
            [smry, pd.DataFrame(index=pd.to_datetime(time_index_dt))], sort=False
        )
        # Drop duplicated dates. It is always the first one which is the
        # original.
        smry = smry[~smry.index.duplicated(keep="first")]

        smry.sort_index(inplace=True)
        smry = smry.apply(pd.to_numeric)

        cummask = smry_cumulative(column_keys)
        cum_columns = [column_keys[i] for i in range(len(column_keys)) if cummask[i]]
        noncum_columns = [
            column_keys[i] for i in range(len(column_keys)) if not cummask[i]
        ]
        if cum_columns:
            smry[cum_columns] = (
                smry[cum_columns]
                .interpolate(method="time")
                .fillna(method="ffill")
                .fillna(method="bfill")
            )
        if noncum_columns:
            smry[noncum_columns] = (
                smry[noncum_columns].fillna(method="bfill").fillna(value=0)
            )

        smry.index = smry.index.set_names(["DATE"])
        return smry.loc[pd.to_datetime(time_index_dt)]

    def get_smry_dates(self, freq="monthly", normalize=False):
        """Return list of datetimes available in the realization

        Similar to the function in ScratchRealization,
        but start and end date is taken from internalized
        smry dataframes.

        Args:
            freq: string denoting requested frequency for
                the list of datetimes.
                'daily', 'monthly', 'yearly' or 'weekly'
                'first' will give out the first date (minimum) and
                'last' will give out the last date (maximum),
                both as lists with one element.
            normalize: Whether to normalize backwards at the start
                and forwards at the end to ensure the entire
                date range is covered.
        Returns:
            list of datetimes. Empty if no summary data is available.
        """
        # Look for available smry, if we have one
        # at the requested frequency, use it directly:
        if shortcut2path(self.keys(), "unsmry--" + freq) in self.keys():
            available_smry = [shortcut2path(self.keys(), "unsmry--" + freq)]
        else:
            # Use all of them.
            available_smry = [x for x in self.keys() if "unsmry" in x]
        if not available_smry:
            raise ValueError("No summary to get start and end date from")

        # Infer start and end-date from internalized smry data
        available_dates = set()
        for smry in available_smry:
            available_dates = available_dates.union(self.get_df(smry)["DATE"].values)

        # Parse every date to datetime, needed?
        available_dates = [pd.to_datetime(x) for x in list(available_dates)]
        start_date = min(available_dates)
        end_date = max(available_dates)
        if freq == "first":
            return [start_date.date()]
        if freq == "last":
            return [end_date.date()]
        if freq in ("custom", "raw"):
            return available_dates
        if normalize:
            raise NotImplementedError
        datetimes = date_range(start_date, end_date, freq=freq)
        # Convert from Pandas' datetime64 to datetime.date:
        return [x.date() for x in datetimes]

    def get_smry_meta(self, column_keys=None):
        """
        Provide metadata for summary data vectors.

        A dictionary indexed by summary vector names is returned, and each
        value is another dictionary with potentially the following metadata types:
        * unit (string)
        * is_total (bool)
        * is_rate (bool)
        * is_historical (bool)
        * get_num (int) (only provided if not None)
        * keyword (str)
        * wgname (str or None)

        Args:
            column_keys (list or str): Column key wildcards.

        Returns:
            dict of dict with metadata information
        """
        # Warning: Code is identical the same function in virtualensemble.py
        if column_keys is None:
            column_keys = ["*"]
        if not isinstance(column_keys, list):
            column_keys = [column_keys]

        available_smrynames = self.get_df("__smry_metadata")["SMRYCOLUMN"].values
        matches = set()
        for key in column_keys:
            matches = matches.union(
                [name for name in available_smrynames if fnmatch.fnmatch(name, key)]
            )
        return (
            self.get_df("__smry_metadata")
            .set_index("SMRYCOLUMN")
            .loc[list(matches), :]
            .to_dict(orient="index")
        )

    def _glob_smry_keys(self, column_keys):
        """Glob a list of column keys

        Given a list of wildcard summary vectors,
        expand the list to the known summary vectors
        (in any internalized smry dataframe)

        Returns empty list if no columns match the
        wildcard(s). This will also happen if there
        is no internalized smry dataframes

        Args:
            column_keys: str or list of str, like
                ['F*PR', 'F*PT']

        Returns:
            list of strings
        """
        if isinstance(column_keys, str):
            column_keys = [column_keys]

        # Get a list ala ['yearly', 'daily']
        available_smry = [x for x in self.keys() if "unsmry" in x]

        if not available_smry:
            raise ValueError(
                "No summary data available. Use load_smry() "
                "before making a virtual realization."
            )

        # Merge all internalized columns:
        available_keys = set()
        for smry in available_smry:
            available_keys = available_keys.union(self.get_df(smry).columns)

        matches = set()
        for key in column_keys:
            matches = matches.union(
                [x for x in available_keys if fnmatch.fnmatch(x, key)]
            )
        if "DATE" in matches:
            matches.remove("DATE")
        return list(matches)

    def __sub__(self, other):
        """Substract another realization from this"""
        result = RealizationCombination(ref=self, sub=other)
        return result

    def __add__(self, other):
        """Add another realization from this"""
        result = RealizationCombination(ref=self, add=other)
        return result

    def __mul__(self, other):
        """Scale this realization by a scalar value"""
        result = RealizationCombination(ref=self, scale=float(other))
        return result

    def __rsub__(self, other):
        """Add another realization from this"""
        result = RealizationCombination(ref=self, sub=other)
        return result

    def __radd__(self, other):
        """Substract another realization from this"""
        result = RealizationCombination(ref=self, add=other)
        return result

    def __rmul__(self, other):
        """Scale this realization by a scalar value"""
        result = RealizationCombination(ref=self, scale=float(other))
        return result

    @property
    def parameters(self):
        """Convenience getter for parameters.txt"""
        return self.data["parameters.txt"]

    @property
    def name(self):
        """Return name of ensemble"""
        return self._description


def smry_cumulative(column_keys):
    """Determine whether smry vectors are cumulative

    Returns list of booleans, indicating whether a certain
    column_key in summary dataframes corresponds to a cumulative
    column.

    The current implementation checks for the letter 'T' in the
    column key, but this behaviour is not guaranteed in the
    future, in case the cumulative information gets internalized

    Warning: This code is duplicated in realization.py, even though
    a ScratchRealization has access to the EclSum object which can
    give the true answer

    Args:
        column_keys: str or list of strings with summary vector
            names
    Returns:
        list of booleans, corresponding to each inputted
            summary vector name.
    """
    if isinstance(column_keys, str):
        column_keys = [column_keys]
    if not isinstance(column_keys, list):
        raise TypeError("column_keys must be str or list of str")
    return [
        (x.endswith("T") and ":" not in x and "CT" not in x)
        or ("T:" in x and "CT:" not in x)
        for x in column_keys
    ]



"""Top-level package for fmu.ensemble"""

try:
    from .version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"

from .ensemble import ScratchEnsemble  # noqa
from .realization import ScratchRealization  # noqa
from .ensembleset import EnsembleSet  # noqa
from .virtualrealization import VirtualRealization  # noqa
from .virtualensemble import VirtualEnsemble  # noqa
from .ensemblecombination import EnsembleCombination  # noqa
from .realizationcombination import RealizationCombination  # noqa
from .observations import Observations  # noqa
